
import asyncio

from aiolimiter import AsyncLimiter

from a_sync import aliases, exceptions
from a_sync._typing import *


@overload
def apply_rate_limit(
    coro_fn: Literal[None],
    runs_per_minute: int,
) -> AsyncDecorator[P, T]:...
    
@overload
def apply_rate_limit(
    coro_fn: int,
    runs_per_minute: Literal[None],
) -> AsyncDecorator[P, T]:...
    
@overload
def apply_rate_limit(
    coro_fn: CoroFn[P, T],
    runs_per_minute: Union[int, AsyncLimiter],
) -> CoroFn[P, T]:...
    
def apply_rate_limit(
    coro_fn: Optional[Union[CoroFn[P, T], int]] = None,
    runs_per_minute: Optional[Union[int, AsyncLimiter]] = None,
) -> AsyncDecoratorOrCoroFn[P, T]:
    # Parse Inputs
    if isinstance(coro_fn, int):
        assert runs_per_minute is None
        runs_per_minute = coro_fn
        coro_fn = None
        
    elif coro_fn is None:
        if runs_per_minute is not None and not isinstance(runs_per_minute, int):
            raise TypeError("'runs_per_minute' must be an integer.", runs_per_minute)
        
    elif not asyncio.iscoroutinefunction(coro_fn):
        raise exceptions.FunctionNotAsync(coro_fn)
    
    def rate_limit_decorator(coro_fn: CoroFn[P, T]) -> CoroFn[P, T]:
        limiter = runs_per_minute if isinstance(runs_per_minute, AsyncLimiter) else AsyncLimiter(runs_per_minute) if runs_per_minute else aliases.dummy
        async def rate_limit_wrap(*args: P.args, **kwargs: P.kwargs) -> T:
            async with limiter:  # type: ignore [attr-defined]
                return await coro_fn(*args, **kwargs)
        return rate_limit_wrap
    
    return rate_limit_decorator if coro_fn is None else rate_limit_decorator(coro_fn)
            
    



import pytest
from time import time
from tests.fixtures import TestSemaphore, increment


# ISSUES
# - We are unable to pass in an existing semaphore object, it attaches to a different loop.
#   Maybe a problem with test suite interaction?
# - semaphore modifier works fine with integer inputs


instance = TestSemaphore(1, sync=False)

@increment
@pytest.mark.asyncio_cooperative
async def test_semaphore(i: int):
    start = time()
    assert await instance.test_fn() == 1
    duration = time() - start
    assert i < 3 or duration > i # There is a 1 second sleep in this fn. If the semaphore is not working, all tests will complete in 1 second.
    
    
@increment
@pytest.mark.asyncio_cooperative
async def test_semaphore_property(i: int):
    start = time()
    assert await instance.test_property == 2
    duration = time() - start
    assert i < 3 or duration > i # There is a 1 second sleep in this fn. If the semaphore is not working, all tests will complete in 1 second.
    
@increment
@pytest.mark.asyncio_cooperative
async def test_semaphore_cached_property(i: int):
    start = time()
    assert await instance.test_cached_property == 3
    duration = time() - start
    # There is a 1 second sleep in this fn but a semaphore override with a value of 50. 
    # You can tell it worked correctly because the class-defined semaphore value is just one, whch would cause this test to fail if it were used.
    # If the override is not working, all tests will complete in just over 1 second.
    assert i == 1 or duration < 1.05 




from a_sync import _flags, config
from a_sync._typing import *
from a_sync.modified import ASyncDecorator, ASyncFunction

########################
# The a_sync decorator #
########################

# @a_sync
# def some_fn():
#     pass
#
# @a_sync
# async def some_fn():
#     pass

@overload
def a_sync(
    **modifiers: Unpack[ModifierKwargs],
) -> AnyFn[P, T]:...

@overload # async def, None default
def a_sync(  # type: ignore [misc]
    coro_fn: CoroFn[P, T],
    default: Literal[None] = None,
    **modifiers: Unpack[ModifierKwargs],
) -> CoroFn[P, T]:...

@overload # sync def none default
def a_sync(  # type: ignore [misc]
    coro_fn: SyncFn[P, T],
    default: Literal[None] = None,
    **modifiers: Unpack[ModifierKwargs],
) -> SyncFn[P, T]:...

# @a_sync(default='async')
# def some_fn():
#     pass
#
# @a_sync(default='async')
# async def some_fn():
#     pass
#
# NOTE These should output a decorator that will be applied to 'some_fn'

@overload 
def a_sync(  # type: ignore [misc]
    coro_fn: Literal[None],
    default: Literal['async'],
    **modifiers: Unpack[ModifierKwargs],
) -> AllToAsyncDecorator[P, T]:...

@overload # if you try to use default as the only arg
def a_sync(  # type: ignore [misc]
    coro_fn: Literal['async'],
    default: Literal[None],
    **modifiers: Unpack[ModifierKwargs],
) -> AllToAsyncDecorator[P, T]:...

# a_sync(some_fn, default='async')

@overload # async def, async default
def a_sync(  # type: ignore [misc]
    coro_fn: CoroFn[P, T],
    default: Literal['async'],
    **modifiers: Unpack[ModifierKwargs],
) -> CoroFn[P, T]:...

@overload # sync def async default
def a_sync(  # type: ignore [misc]
    coro_fn: SyncFn[P, T],
    default: Literal['async'],
    **modifiers: Unpack[ModifierKwargs],
) -> CoroFn[P, T]:...

# a_sync(some_fn, default='sync')

@overload # async def, sync default
def a_sync(  # type: ignore [misc]
    coro_fn: CoroFn[P, T],
    default: Literal['sync'],
    **modifiers: Unpack[ModifierKwargs],
) -> SyncFn[P, T]:...

@overload # sync def sync default
def a_sync(  # type: ignore [misc]
    coro_fn: SyncFn[P, T],
    default: Literal['sync'],
    **modifiers: Unpack[ModifierKwargs],
) -> SyncFn[P, T]:...

# @a_sync(default='sync')
# def some_fn():
#     pass
#
# @a_sync(default='sync')
# async def some_fn():
#     pass
#
# NOTE These should output a decorator that will be applied to 'some_fn'

@overload 
def a_sync(  # type: ignore [misc]
    coro_fn: Literal[None],
    default: Literal['sync'],
    **modifiers: Unpack[ModifierKwargs],
) -> AllToSyncDecorator[P, T]:...

@overload # if you try to use default as the only arg
def a_sync(  # type: ignore [misc]
    coro_fn: Literal['sync'],
    default: Literal[None] = None,
    **modifiers: Unpack[ModifierKwargs],
) -> AllToSyncDecorator[P, T]:...

@overload # if you try to use default as the only arg
def a_sync(  # type: ignore [misc]
    coro_fn: Literal['sync'],
    default: Literal[None],
    **modifiers: Unpack[ModifierKwargs],
) -> AllToSyncDecorator[P, T]:...
    
# catchall
def a_sync(  # type: ignore [misc]
    coro_fn: Optional[AnyFn[P, T]] = None,
    default: DefaultMode = config.DEFAULT_MODE,
    **modifiers: Unpack[ModifierKwargs],  # default values are set by passing these kwargs into a ModifierManager object.
) -> Union[ASyncDecorator[P, T], ASyncFunction[P, T]]:
    f"""
    A coroutine function decorated with this decorator can be called as a sync function by passing a boolean value for any of these kwargs: {_flags.VIABLE_FLAGS}
    
    lib defaults:
    async settings
        cache_type: CacheType = None,            - This can be None or 'memory'. 'memory' is a lru cache which can be modified with the 'cache_typed','ram_cache_maxsize','ram_cache_ttl' modifiers.
        cache_typed: bool = False,               - Set to True if you want types considered treated for cache keys. ie with cache_typed=True, Decimal(0) and 0 will be considered separate keys.
        ram_cache_maxsize: Optional[int] = -1,   - The maxsize for your lru cache. None if cache is unbounded. If you set this value without specifying a cache type, 'memory' will automatically be applied.
        ram_cache_ttl: Optional[int] = None,     - The ttl for items in your lru cache. Set to None. If you set this value without specifying a cache type, 'memory' will automatically be applied.
        runs_per_minute: Optional[int] = None,   - Setting this value enables a rate limiter for the decorated function.
        semaphore: SemaphoreSpec = None,         - drop in a Semaphore for your async defined functions.
    sync settings
        executor: Executor = config.default_sync_executor

    ######################
    # async def examples #
    ######################

    @a_sync
    async def some_fn():
        reurn True
    
    await some_fn() == True
    some_fn(sync=True) == True

    
    @a_sync(default=None)
    async def some_fn():
        reurn True
    
    await some_fn() == True
    some_fn(sync=True) == True

    
    @a_sync(default='async')
    async def some_fn():
        reurn True
    
    await some_fn() == True
    some_fn(sync=True) == True
    

    @a_sync(default='sync')
    async def some_fn():
        reurn True
    
    some_fn() == True
    await some_fn(sync=False) == True

    
    # TODO: in a future release, I will make this usable with sync functions as well

    ################
    # def examples #
    ################

    @a_sync
    def some_fn():
        reurn True
    
    some_fn() == True
    await some_fn(sync=False) == True
    

    @a_sync(default=None)
    def some_fn():
        reurn True
    
    some_fn() == True
    await some_fn(sync=False) == True


    @a_sync(default='async')
    def some_fn():
        reurn True
    
    await some_fn() == True
    some_fn(sync=True) == True


    @a_sync(default='sync')
    def some_fn():
        reurn True
    
    some_fn() == True
    await some_fn(sync=False) == True
    """
    
    # If the dev tried passing a default as an arg instead of a kwarg, ie: @a_sync('sync')...
    if coro_fn in ['async', 'sync']:
        default = coro_fn  # type: ignore [assignment]
        coro_fn = None
    
    deco = ASyncDecorator(default=default, **modifiers)
    return deco if coro_fn is None else deco(coro_fn)  # type: ignore [arg-type]




import abc
from typing import Union

from a_sync import _flags, _kwargs, exceptions, modifiers
from a_sync._meta import ASyncMeta
from a_sync._typing import *


class ASyncABC(metaclass=ASyncMeta):

    ##################################
    # Concrete Methods (overridable) #
    ##################################

    def __a_sync_should_await__(self, kwargs: dict, force: Optional[Literal[True]] = None) -> bool:
        """Returns a boolean that indicates whether methods of 'instance' should be called as sync or async methods."""
        try:
            # Defer to kwargs always
            return self.__should_await_from_kwargs(kwargs)
        except exceptions.NoFlagsFound:
            # No flag found in kwargs, check for a flag attribute.
            return force if force else self.__should_await_from_instance

    @property
    def __should_await_from_instance(self) -> bool:
        """You can override this if you want."""
        return _flags.negate_if_necessary(self.__a_sync_flag_name__, self.__a_sync_flag_value__)
    
    def __should_await_from_kwargs(self, kwargs: dict) -> bool:
        """You can override this if you want."""
        return _kwargs.is_sync(kwargs, pop_flag=True)
    
    @classmethod
    def __a_sync_instance_will_be_sync__(cls, kwargs: dict) -> bool:
        """You can override this if you want."""
        try:
            return _kwargs.is_sync(kwargs)
        except exceptions.NoFlagsFound:
            return cls.__a_sync_default_mode__()  # type: ignore [return-value]

    ######################################
    # Concrete Methods (non-overridable) #
    ######################################
    
    @property
    def __a_sync_modifiers__(self: "ASyncABC") -> ModifierKwargs:
        """You should not override this."""
        return modifiers.get_modifiers_from(self)

    ####################
    # Abstract Methods #
    ####################

    @abc.abstractproperty
    def __a_sync_flag_name__(self) -> str:
        pass
    
    @abc.abstractproperty
    def __a_sync_flag_value__(self) -> bool:
        pass
    
    @abc.abstractclassmethod  # type: ignore [misc]
    def __a_sync_default_mode__(cls) -> bool:
        ...


import asyncio
import functools
from collections import defaultdict
from threading import Thread, current_thread

from a_sync import exceptions
from a_sync._typing import *


class ThreadsafeSemaphore(asyncio.Semaphore):
    """
    While its a bit weird to run multiple event loops, sometimes either you or a lib you're using must do so. 
    When in use in threaded applications, this semaphore will not work as intended but at least your program will function.
    You may need to reduce the semaphore value for multi-threaded applications.
    
    # TL;DR it's a janky fix for an edge case problem and will otherwise function as a normal asyncio.Semaphore.
    """

    def __init__(self, value: Optional[int]) -> None:
        assert isinstance(value, int), f"{value} should be an integer."
        self._value = value
        self.semaphores: DefaultDict[Thread, asyncio.Semaphore] = defaultdict(lambda: asyncio.Semaphore(value))  # type: ignore [arg-type]
        self.dummy = DummySemaphore()
    
    def __repr__(self) -> str:
        return f"<ThreadsafeSemaphore value={self._value}>"
    
    @property
    def use_dummy(self) -> bool:
        return self._value is None
    
    @property
    def semaphore(self) -> asyncio.Semaphore:
        if self.use_dummy:
            return self.dummy
        tid = current_thread()
        if tid not in self.semaphores:
            self.semaphores[tid] = asyncio.Semaphore(self._value)
        return self.semaphores[tid]
    
    async def __aenter__(self):
        await self.semaphore.acquire()
    
    async def __aexit__(self, *args):
        self.semaphore.release()


class DummySemaphore(asyncio.Semaphore):
    def __init__(*args, **kwargs):
        ...
    def __repr__(self) -> str:
        return "<DummySemaphore>"
    async def __aenter__(self):
        ...
    async def __aexit__(self, *args):
        ...


Semaphore = Union[
    asyncio.Semaphore,
    asyncio.BoundedSemaphore,
    ThreadsafeSemaphore,
    DummySemaphore,
]

@overload
async def apply_semaphore(  # type: ignore [misc]
    coro_fn: Literal[None],
    semaphore: SemaphoreSpec,
) -> AsyncDecorator[P, T]:...

@overload
async def apply_semaphore(
    coro_fn: SemaphoreSpec,
    semaphore: Literal[None],
) -> AsyncDecorator[P, T]:...

@overload
async def apply_semaphore(
    coro_fn: CoroFn[P, T],
    semaphore: SemaphoreSpec,
) -> CoroFn[P, T]:...
    
def apply_semaphore(
    coro_fn: Optional[Union[CoroFn[P, T], SemaphoreSpec]] = None,
    semaphore: SemaphoreSpec = None,
) -> AsyncDecoratorOrCoroFn[P, T]:
    # Parse Inputs
    if isinstance(coro_fn, (int, asyncio.Semaphore)):
        if semaphore is not None:
            raise ValueError("You can only pass in one arg.")
        semaphore = coro_fn
        coro_fn = None
        
    elif not asyncio.iscoroutinefunction(coro_fn):
        raise exceptions.FunctionNotAsync(coro_fn)
        
    # Create the semaphore if necessary
    if isinstance(semaphore, int):
        semaphore = ThreadsafeSemaphore(semaphore)
    elif not isinstance(semaphore, asyncio.Semaphore):
        raise TypeError(f"'semaphore' must either be an integer or a Semaphore object.")
        
    # Create and return the decorator
    def semaphore_decorator(coro_fn: CoroFn[P, T]) -> CoroFn[P, T]:
        @functools.wraps(coro_fn)
        async def semaphore_wrap(*args, **kwargs) -> T:
            async with semaphore:  # type: ignore [union-attr]
                return await coro_fn(*args, **kwargs)
        return semaphore_wrap
    return semaphore_decorator if coro_fn is None else semaphore_decorator(coro_fn)


dummy_semaphore = DummySemaphore()




import asyncio
from concurrent.futures._base import Executor
from typing import (TYPE_CHECKING, Any, Awaitable, Callable, DefaultDict, Dict,
                    Generic, Iterator, List, Literal, Optional, Tuple,
                    Type, TypedDict, TypeVar, Union, overload)

from typing_extensions import Concatenate, ParamSpec, Unpack

if TYPE_CHECKING:
    from a_sync.abstract import ASyncABC

T = TypeVar("T")
P = ParamSpec("P")

MaybeAwaitable = Union[Awaitable[T], T]

Property = Callable[["ASyncABC"], T]
HiddenMethod = Callable[["ASyncABC", Dict[str, bool]], T]
AsyncBoundMethod = Callable[Concatenate["ASyncABC", P], Awaitable[T]]
BoundMethod = Callable[Concatenate["ASyncABC", P], T]

CoroFn = Callable[P, Awaitable[T]]
SyncFn = Callable[P, T]
AnyFn = Union[CoroFn[P, T], SyncFn[P, T]]

AsyncDecorator = Callable[[CoroFn[P, T]], CoroFn[P, T]]

AllToAsyncDecorator = Callable[[AnyFn[P, T]], CoroFn[P, T]]
AllToSyncDecorator = Callable[[AnyFn[P, T]], SyncFn[P, T]]

AsyncDecoratorOrCoroFn = Union[AsyncDecorator[P, T], CoroFn[P, T]]

DefaultMode = Literal['sync', 'async', None]

CacheType = Literal['memory', None]
SemaphoreSpec = Optional[Union[asyncio.Semaphore, int]]

class ModifierKwargs(TypedDict, total=False):
    default: DefaultMode
    cache_type: CacheType
    cache_typed: bool
    ram_cache_maxsize: Optional[int]
    ram_cache_ttl: Optional[int]
    runs_per_minute: Optional[int]
    semaphore: SemaphoreSpec
    # sync modifiers
    executor: Executor




from a_sync import aliases
from a_sync.base import ASyncGenericBase
from a_sync.decorator import a_sync
from a_sync.semaphores import (DummySemaphore, ThreadsafeSemaphore,
                               apply_semaphore)
from a_sync.singleton import ASyncGenericSingleton

# I alias the aliases for your convenience.
# I prefer "aka" but its meaning is not intuitive when reading code so I created both aliases for you to choose from.
# NOTE: Overkill? Maybe.
aka = alias = aliases

# alias for backward-compatability, will be removed eventually, probably in 0.1.0
ASyncBase = ASyncGenericBase    





import inspect

from a_sync import _flags, exceptions
from a_sync.abstract import ASyncABC


class ASyncGenericBase(ASyncABC):
    """
    Inherit from this class to a-syncify all of your bound methods.
    Allows for the use of a variety of flags out-of-box.
    You can choose which flag(s) work best for your subclass implementation.
    """
    @property
    def __a_sync_flag_name__(self) -> str:
        try:
            return self.__get_a_sync_flag_name_from_signature()
        except exceptions.ASyncFlagException:
            # We can't get the flag name from the __init__ signature,
            # but maybe the implementation sets the flag somewhere else.
            # Let's check the instance's atributes
            present_flags = [flag for flag in _flags.VIABLE_FLAGS if hasattr(self, flag)]
            if len(present_flags) == 0:
                raise exceptions.NoFlagsFound(self)
            if len(present_flags) > 1:
                raise exceptions.TooManyFlags(self, present_flags)
            return present_flags[0]
        
    @property
    def __a_sync_flag_value__(self) -> bool:
        flag = self.__a_sync_flag_name__
        if not isinstance(flag, str):
            raise TypeError(flag, type(flag))
        flag_value = getattr(self, flag)
        if not isinstance(flag_value, bool):
            raise exceptions.InvalidFlagValue(flag, flag_value)
        return flag_value

    @classmethod  # type: ignore [misc]
    def __a_sync_default_mode__(cls) -> bool:
        flag = cls.__get_a_sync_flag_name_from_signature()
        flag_value = cls.__a_sync_flag_default_value_from_signature
        return _flags.negate_if_necessary(flag, flag_value)  # type: ignore [arg-type]
    
    @classmethod
    #@property # TODO debug why making this into a classproperty breaks it
    def __get_a_sync_flag_name_from_signature(cls) -> str:
        if cls.__name__ == "ASyncGenericBase":
            return None
        signature = inspect.signature(cls.__init__)
        present_flags = [flag for flag in _flags.VIABLE_FLAGS if flag in signature.parameters]
        if len(present_flags) == 0:
            raise exceptions.NoFlagsFound(cls, signature.parameters.keys())
        if len(present_flags) > 1:
            raise exceptions.TooManyFlags(cls, present_flags)
        return present_flags[0]

    @classmethod  # type: ignore [misc]
    @property
    def __a_sync_flag_default_value_from_signature(cls) -> bool:
        signature = inspect.signature(cls.__init__)
        flag = cls.__get_a_sync_flag_name_from_signature()
        flag_value = signature.parameters[flag].default
        if flag_value is inspect._empty:  # type: ignore [attr-defined]
            raise Exception("The implementation for 'cls' uses an arg to specify sync mode, instead of a kwarg. We are unable to proceed. I suppose we can extend the code to accept positional arg flags if necessary")
        return _flags.validate_flag_value(flag, flag_value)




import asyncio

from async_lru import alru_cache

from a_sync import exceptions
from a_sync._typing import *

class CacheKwargs(TypedDict):
    maxsize: Optional[int]
    ttl: Optional[int]
    typed: bool

@overload
def apply_async_memory_cache(
    coro_fn: Literal[None],
    **kwargs: Unpack[CacheKwargs]
) -> AsyncDecorator[P, T]:...
    
@overload
def apply_async_memory_cache(
    coro_fn: int,
    **kwargs: Unpack[CacheKwargs]
) -> AsyncDecorator[P, T]:...
    
@overload
def apply_async_memory_cache(
    coro_fn: CoroFn[P, T],
    **kwargs: Unpack[CacheKwargs]
) -> CoroFn[P, T]:...

@overload
def apply_async_memory_cache(
    coro_fn: Literal[None],
    **kwargs: Unpack[CacheKwargs]
) -> AsyncDecorator[P, T]:...

def apply_async_memory_cache(
    coro_fn: Optional[Union[CoroFn[P, T], int]] = None,
    maxsize: Optional[int] = None,
    ttl: Optional[int] = None,
    typed: bool = False,
) -> AsyncDecoratorOrCoroFn[P, T]:
    # Parse Inputs
    if isinstance(coro_fn, int):
        assert maxsize is None
        maxsize = coro_fn
        coro_fn = None
    
    # Validate 
    elif coro_fn is None:
        if maxsize is not None and not isinstance(maxsize, int):
            raise TypeError("'lru_cache_maxsize' must be an integer or None.", maxsize)
    elif not asyncio.iscoroutinefunction(coro_fn):
        raise exceptions.FunctionNotAsync(coro_fn)

    cache_decorator = alru_cache(maxsize=maxsize, ttl=ttl, typed=typed)
    return cache_decorator if coro_fn is None else cache_decorator(coro_fn)




import asyncio
import pytest
from typing import Literal

import a_sync


def _test_kwargs(fn, default: Literal['sync','async',None]):
    # force async
    assert asyncio.get_event_loop().run_until_complete(fn(sync=False)) == 2
    assert asyncio.get_event_loop().run_until_complete(fn(asynchronous=True)) == 2
    # force sync
    with pytest.raises(TypeError):
        assert asyncio.get_event_loop().run_until_complete(fn(sync=True)) == 2
    with pytest.raises(TypeError):
        assert asyncio.get_event_loop().run_until_complete(fn(asynchronous=False)) == 2
    assert fn(sync=True) == 2
    assert fn(asynchronous=False) == 2
    if default == 'sync':
        assert fn() == 2
    elif default == 'async':
        assert asyncio.get_event_loop().run_until_complete(fn()) == 2

# ASYNC DEF
def test_decorator_no_args():
    @a_sync.a_sync
    async def some_test_fn() -> int:
        return 2
    assert asyncio.get_event_loop().run_until_complete(some_test_fn()) == 2
    _test_kwargs(some_test_fn, None)
    
    @a_sync.a_sync()
    async def some_test_fn() -> int:
        return 2
    assert asyncio.get_event_loop().run_until_complete(some_test_fn()) == 2
    _test_kwargs(some_test_fn, None)
    
def test_decorator_default_none_arg():
    @a_sync.a_sync(None)
    async def some_test_fn() -> int:
        return 2
    assert asyncio.get_event_loop().run_until_complete(some_test_fn()) == 2
    _test_kwargs(some_test_fn, None)
    
def test_decorator_default_none_kwarg():
    @a_sync.a_sync(default=None)
    async def some_test_fn() -> int:
        return 2
    assert asyncio.get_event_loop().run_until_complete(some_test_fn()) == 2
    _test_kwargs(some_test_fn, None)

def test_decorator_default_sync_arg():
    @a_sync.a_sync('sync')
    async def some_test_fn() -> int:
        return 2
    with pytest.raises(TypeError):
        asyncio.get_event_loop().run_until_complete(some_test_fn())
    assert some_test_fn() == 2
    _test_kwargs(some_test_fn, 'sync')

def test_decorator_default_sync_kwarg():
    @a_sync.a_sync(default='sync')
    async def some_test_fn() -> int:
        return 2
    with pytest.raises(TypeError):
        asyncio.get_event_loop().run_until_complete(some_test_fn())
    assert some_test_fn() == 2
    _test_kwargs(some_test_fn, 'sync')
    
def test_decorator_default_async_arg():
    @a_sync.a_sync('async')
    async def some_test_fn() -> int:
        return 2
    assert asyncio.get_event_loop().run_until_complete(some_test_fn()) == 2
    _test_kwargs(some_test_fn, 'async')
    
def test_decorator_default_async_kwarg():
    @a_sync.a_sync(default='async')
    async def some_test_fn() -> int:
        return 2
    assert asyncio.get_event_loop().run_until_complete(some_test_fn()) == 2
    _test_kwargs(some_test_fn, 'async')


# SYNC DEF
def test_sync_decorator_no_args():
    @a_sync.a_sync
    def some_test_fn() -> int:
        return 2
    assert some_test_fn() == 2
    _test_kwargs(some_test_fn, None)
    
    @a_sync.a_sync()
    def some_test_fn() -> int:
        return 2
    assert some_test_fn() == 2
    _test_kwargs(some_test_fn, None)
    
def test_sync_decorator_default_none_arg():
    @a_sync.a_sync(None)
    def some_test_fn() -> int:
        return 2
    assert some_test_fn() == 2
    _test_kwargs(some_test_fn, None)
    
def test_sync_decorator_default_none_kwarg():
    @a_sync.a_sync(default=None)
    def some_test_fn() -> int:
        return 2
    assert some_test_fn() == 2
    _test_kwargs(some_test_fn, None)

def test_sync_decorator_default_sync_arg():
    @a_sync.a_sync('sync')
    def some_test_fn() -> int:
        return 2
    assert some_test_fn() == 2
    _test_kwargs(some_test_fn, 'sync')

def test_sync_decorator_default_sync_kwarg():
    @a_sync.a_sync(default='sync')
    def some_test_fn() -> int:
        return 2
    assert some_test_fn() == 2
    _test_kwargs(some_test_fn, 'sync')
    
def test_sync_decorator_default_async_arg():
    @a_sync.a_sync('async')
    def some_test_fn() -> int:
        return 2
    assert asyncio.get_event_loop().run_until_complete(some_test_fn()) == 2
    _test_kwargs(some_test_fn, 'async')
    
def test_sync_decorator_default_async_kwarg():
    @a_sync.a_sync(default='async')
    def some_test_fn() -> int:
        return 2
    assert asyncio.get_event_loop().run_until_complete(some_test_fn()) == 2
    _test_kwargs(some_test_fn, 'async')




from a_sync import _flags, exceptions


def get_flag_name(kwargs: dict) -> str:
    present_flags = [flag for flag in _flags.VIABLE_FLAGS if flag in kwargs]
    if len(present_flags) == 0:
        raise exceptions.NoFlagsFound('kwargs', kwargs.keys())
    if len(present_flags) != 1:
        raise exceptions.TooManyFlags('kwargs', present_flags)
    return present_flags[0]

def is_sync(kwargs: dict, pop_flag: bool = False) -> bool:
    flag = get_flag_name(kwargs)
    flag_value = kwargs.pop(flag) if pop_flag else kwargs[flag]
    return _flags.negate_if_necessary(flag, flag_value)



import functools
import os
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor
from concurrent.futures._base import Executor

from a_sync._typing import *

EXECUTOR_TYPE = os.environ.get("A_SYNC_EXECUTOR_TYPE", "threads")
EXECUTOR_VALUE = int(os.environ.get("A_SYNC_EXECUTOR_VALUE", 8))

@functools.lru_cache(maxsize=1)
def get_default_executor() -> Executor:
    if EXECUTOR_TYPE.lower().startswith('p'): # p, P, proc, Processes, etc
        return ProcessPoolExecutor(EXECUTOR_VALUE)
    elif EXECUTOR_TYPE.lower().startswith('t'): # t, T, thread, THREADS, etc
        return ThreadPoolExecutor(EXECUTOR_VALUE)
    raise ValueError("Invalid value for A_SYNC_EXECUTOR_TYPE. Please use 'threads' or 'processes'.")

default_sync_executor = get_default_executor()

null_modifiers = ModifierKwargs(
    default=None,
    cache_type=None,
    cache_typed=False,
    ram_cache_maxsize=-1,
    ram_cache_ttl=None,
    runs_per_minute=None,
    semaphore=None,
    executor=default_sync_executor,
)

#####################
# Default Modifiers #
#####################

# User configurable default modifiers to be applied to any a_sync decorated function if you do not specify kwarg values for each modifier.

DEFAULT_MODE: DefaultMode = os.environ.get("A_SYNC_DEFAULT_MODE")  # type: ignore [assignment]
CACHE_TYPE: CacheType = typ if (typ := os.environ.get("A_SYNC_CACHE_TYPE", "").lower()) else null_modifiers['cache_type']
CACHE_TYPED = bool(os.environ.get("A_SYNC_CACHE_TYPED"))
RAM_CACHE_MAXSIZE = int(os.environ.get("A_SYNC_RAM_CACHE_MAXSIZE", -1)) 
RAM_CACHE_TTL = ttl if (ttl := float(os.environ.get("A_SYNC_RAM_CACHE_TTL", 0))) else null_modifiers['ram_cache_ttl']

RUNS_PER_MINUTE = rpm if (rpm := int(os.environ.get("A_SYNC_RUNS_PER_MINUTE", 0))) else null_modifiers['runs_per_minute']
SEMAPHORE = rpm if (rpm := int(os.environ.get("A_SYNC_SEMAPHORE", 0))) else null_modifiers['semaphore']

user_set_default_modifiers = ModifierKwargs(
    default=DEFAULT_MODE,
    cache_type=CACHE_TYPE,
    cache_typed=CACHE_TYPED,
    ram_cache_maxsize=RAM_CACHE_MAXSIZE,
    ram_cache_ttl=RAM_CACHE_TTL,
    runs_per_minute=RUNS_PER_MINUTE,
    semaphore=SEMAPHORE,
    executor=default_sync_executor,
)




import threading
from abc import ABCMeta
from typing import Any, Dict, Tuple

from a_sync import _bound, modifiers
from a_sync.modified import ASyncFunction, Modified
from a_sync.property import PropertyDescriptor


class ASyncMeta(ABCMeta):
    """Any class with metaclass ASyncMeta will have its functions wrapped with a_sync upon class instantiation."""
    def __new__(cls, name, bases, attrs):
        # Wrap all methods with a_sync abilities
        for attr_name, attr_value in list(attrs.items()):
            # Read modifiers from class definition 
            # NOTE: Open uesion: what do we do when a parent class and subclass define the same modifier differently?
            #       Currently the parent value is used for functions defined on the parent, 
            #       and the subclass value is used for functions defined on the subclass.
            fn_modifiers = modifiers.get_modifiers_from(attrs)
            # Special handling for functions decorated with a_sync decorators
            if isinstance(attr_value, Modified):
                # Check for modifier overrides defined on the Modified object
                fn_modifiers.update(attr_value.modifiers._modifiers)
                if isinstance(attr_value, PropertyDescriptor):
                    # Wrap property
                    attrs[attr_name], attrs[attr_value.hidden_method_name] = _bound._wrap_property(attr_value, **fn_modifiers)
                elif isinstance(attr_value, ASyncFunction):
                    attrs[attr_name] = _bound._wrap_bound_method(attr_value, **fn_modifiers)
                else:
                    raise NotImplementedError(attr_name, attr_value)
                    
            if callable(attr_value) and "__" not in attr_name and not attr_name.startswith("_"):
                # NOTE We will need to improve this logic if somebody needs to use it with classmethods or staticmethods.
                attrs[attr_name] = _bound._wrap_bound_method(attr_value, **fn_modifiers)
        return super(ASyncMeta, cls).__new__(cls, name, bases, attrs)    


class ASyncSingletonMeta(ASyncMeta):
    def __init__(cls, name: str, bases: Tuple[type, ...], namespace: Dict[str, Any]) -> None:
        cls.__instances: Dict[bool, object] = {}
        cls.__lock = threading.Lock()
        super().__init__(name, bases, namespace)

    def __call__(cls, *args: Any, **kwargs: Any):
        is_sync = cls.__a_sync_instance_will_be_sync__(kwargs)  # type: ignore [attr-defined]
        if is_sync not in cls.__instances:
            with cls.__lock:
                # Check again in case `__instance` was set while we were waiting for the lock.
                if is_sync not in cls.__instances:
                    cls.__instances[is_sync] = super().__call__(*args, **kwargs)
        return cls.__instances[is_sync]
